/*****************************************************************************/
/*  local.c - functions that will use the local system                       */
/*  Copyright (C) 1998-2000 Brian Masney <masneyb@seul.org>                  */
/*                                                                           */
/*  This library is free software; you can redistribute it and/or            */
/*  modify it under the terms of the GNU Library General Public              */
/*  License as published by the Free Software Foundation; either             */
/*  version 2 of the License, or (at your option) any later version.         */
/*                                                                           */
/*  This library is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU        */
/*  Library General Public License for more details.                         */
/*                                                                           */
/*  You should have received a copy of the GNU Library General Public        */
/*  License along with this library; if not, write to the Free Software      */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "protocols.h"

static int local_connect 			( gftp_request * request );
static void local_disconnect 			( gftp_request * request );
static long local_get_file 			( gftp_request * request, 
						  const char *filename,
						  long startsize );
static int local_put_file 			( gftp_request * request, 
						  const char *filename,
						  long startsize,
						  long totalsize );
static int local_end_transfer 			( gftp_request * request );
static int local_list_files 			( gftp_request * request );
static long local_get_file_size 		( gftp_request * request,
						  const char *filename );
static int local_chdir 				( gftp_request * request, 
						  const char *directory );
static int local_rmdir 				( gftp_request * request, 
						  const char *directory );
static int local_rmfile 			( gftp_request * request, 
						  const char *file );
static int local_mkdir 				( gftp_request * request, 
						  const char *directory );
static int local_rename 			( gftp_request * request, 
						  const char *oldname,
						  const char *newname );
static int local_chmod 				( gftp_request * request, 
						  const char *file, 
						  int mode );
static int local_set_file_time 			( gftp_request * request, 
						  const char *file,
						  time_t datetime );

void
local_init (gftp_request * request)
{
  g_return_if_fail (request != NULL);

  request->init = local_init;
  request->connect = local_connect;
  request->disconnect = local_disconnect;
  request->get_file = local_get_file;
  request->put_file = local_put_file;
  request->transfer_file = NULL;
  request->get_next_file_chunk = NULL;
  request->put_next_file_chunk = NULL;
  request->end_transfer = local_end_transfer;
  request->list_files = local_list_files;
  request->get_next_file = rfc959_get_next_file;
  request->set_data_type = NULL;
  request->get_file_size = local_get_file_size;
  request->chdir = local_chdir;
  request->rmdir = local_rmdir;
  request->rmfile = local_rmfile;
  request->mkdir = local_mkdir;
  request->rename = local_rename;
  request->chmod = local_chmod;
  request->set_file_time = local_set_file_time;
  request->site = NULL;
  request->url_prefix = "file";
  request->protocol_name = "Local";
  request->unsafe_symlinks = 0;
  request->need_hostport = 0;
  request->need_userpass = 0;
  request->use_cache = 0;
  request->use_threads = 0;
}


static int
local_connect (gftp_request * request)
{
  char tempstr[PATH_MAX];

  g_return_val_if_fail (request != NULL, -2);

  request->sockfd = (void *) 1;

  if (request->directory)
    {
      chdir (request->directory);
      g_free (request->directory);
      request->directory = NULL;
    }
  if (getcwd (tempstr, sizeof (tempstr)) != NULL)
    {
      tempstr[sizeof (tempstr) - 1] = '\0';
      request->directory = g_malloc (strlen (tempstr) + 1);
      strcpy (request->directory, tempstr);
    }

  return (0);
}


static void
local_disconnect (gftp_request * request)
{
  g_return_if_fail (request != NULL);

  request->sockfd = NULL;
  return;
}


static long
local_get_file (gftp_request * request, const char *filename, long startsize)
{
  long size;

  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (filename != NULL, -2);

  if ((request->datafd = fopen (filename, "rb")) == NULL)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
				   _
				   ("Error: Cannot open local file %s: %s\n"),
				   filename, g_strerror (errno));
      return (-2);
    }

  lseek (fileno (request->datafd), 0, SEEK_END);
  size = ftell (request->datafd);
  lseek (fileno (request->datafd), startsize, SEEK_SET);
  return (size);
}


static int
local_put_file (gftp_request * request, const char *filename, long startsize,
                long totalsize)
{
  int sock;

  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (filename != NULL, -2);

  if ((sock = open (filename, 
      startsize > 0 ? O_WRONLY | O_APPEND | O_CREAT: O_WRONLY | O_CREAT, 
      S_IRUSR | S_IWUSR)) < 0)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
				   _("Error: Cannot open local file %s: %s\n"),
				   filename, g_strerror (errno));
      return (-2);
    }

  request->datafd = fdopen (sock, "wb");
  fseek (request->datafd, startsize, SEEK_SET);
  return (0);
}


static int
local_end_transfer (gftp_request * request)
{
  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (request->datafd != NULL, -2);

  fclose (request->datafd);
  request->datafd = NULL;
  return (0);
}


static int
local_list_files (gftp_request * request)
{
  g_return_val_if_fail (request != NULL, -2);

  if (chdir (request->directory) != 0)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
			_("Could not change local directory to %s: %s\n"),
			request->directory, g_strerror (errno));
      return (-2);
    }

  if ((request->datafd = popen ("/bin/ls -al", "r")) == NULL)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
			   _("Could not get local directory listing %s: %s\n"),
			   request->directory, g_strerror (errno));
      return (-1);
    }

  return (request->datafd == NULL);
}


static long
local_get_file_size (gftp_request * request, const char *filename)
{
  struct stat st;

  if (stat (filename, &st) == -1)
    return (-1);
  return (st.st_size);
}


static int
local_chdir (gftp_request * request, const char *directory)
{
  char tempstr[255];

  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (directory != NULL, -2);

  if (chdir (directory) == 0)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
			_("Successfully changed local directory to %s\n"),
			directory);
      if (getcwd (tempstr, sizeof (tempstr)) == NULL)
	{
	  if (request->logging)
	    request->logging_function (gftp_logging_error, request->user_data,
			     _("Could not get current working directory: %s\n"),
			     g_strerror (errno));
	  return (-1);
	}
      if (request->directory)
	g_free (request->directory);
      request->directory = g_malloc (strlen (tempstr) + 1);
      strcpy (request->directory, tempstr);
      return (0);
    }
  if (request->logging)
    request->logging_function (gftp_logging_error, request->user_data,
			    _("Could not change local directory to %s: %s\n"),
			    directory, g_strerror (errno));
  return (-1);
}


static int
local_rmdir (gftp_request * request, const char *directory)
{
  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (directory != NULL, -2);

  if (rmdir (directory) == 0)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
				   _("Successfully removed %s\n"), directory);
      return (0);
    }
  else
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
			   _("Error: Could not remove directory %s: %s\n"),
			   directory, g_strerror (errno));
      return (-1);
    }
}


static int
local_rmfile (gftp_request * request, const char *file)
{
  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (file != NULL, -2);

  if (unlink (file) == 0)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
				   _("Successfully removed %s\n"), file);
      return (0);
    }
  else
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
				   _("Error: Could not remove file %s: %s\n"),
				   file, g_strerror (errno));
      return (-1);
    }
}


static int
local_mkdir (gftp_request * request, const char *directory)
{
  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (directory != NULL, -2);

  if (mkdir (directory, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == 0)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
				   _("Successfully made directory %s\n"),
				   directory);
      return (0);
    }
  else
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
			   _("Error: Could not make directory %s: %s\n"),
			   directory, g_strerror (errno));
      return (-1);
    }
}


static int
local_rename (gftp_request * request, const char *oldname,
	      const char *newname)
{
  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (oldname != NULL, -2);
  g_return_val_if_fail (newname != NULL, -2);

  if (rename (oldname, newname) == 0)
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
				   _("Successfully renamed %s to %s\n"),
				   oldname, newname);
      return (0);
    }
  else
    {
      if (request->logging)
	request->logging_function (gftp_logging_error, request->user_data,
			   _("Error: Could not rename %s to %s: %s\n"),
			   oldname, newname, g_strerror (errno));
      return (-1);
    }
}


static int
local_chmod (gftp_request * request, const char *file, int mode)
{
  char buf[10];
  int newmode;

  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (file != NULL, -2);

  g_snprintf (buf, sizeof (buf), "%d", mode);
  newmode = strtol (buf, NULL, 8);

  if (chmod (file, newmode) == 0) 
    {
      if (request->logging) 
        request->logging_function (gftp_logging_error, request->user_data, 
                                  _("Successfully changed mode of %s to %d\n"),
                                  file, mode);
      return (0);
    }
  else 
    {
      if (request->logging)
        request->logging_function (gftp_logging_error, request->user_data, 
                           _("Error: Could not change mode of %s to %d: %s\n"),
                           file, mode, g_strerror (errno));
      return (-1);
    }
}


static int
local_set_file_time (gftp_request * request, const char *file,
		     time_t datetime)
{
  struct utimbuf time_buf;

  g_return_val_if_fail (request != NULL, -2);
  g_return_val_if_fail (file != NULL, -2);

  time_buf.modtime = time_buf.actime = datetime;
  return (utime (file, &time_buf));
}

